from typing import List

from guet.files import File, FileSystem
from guet.util import project_root

from .hook import shared_hook_lines

_FILE_NAMES = [
    'pre-commit',
    'post-commit',
    'commit-msg',
    'pre-commit-guet',
    'post-commit-guet',
    'commit-msg-guet',
]

def _is_guet_file(content: List[str]) -> bool:
    return content[1:] == shared_hook_lines()


def all_guet_hooks(file_system: FileSystem) -> List[File]:
    git_hooks_path = project_root().joinpath('.git').joinpath('hooks')
    found = []
    for file_name in _FILE_NAMES:
        file = file_system.get(git_hooks_path.joinpath(file_name))
        if _is_guet_file(file.read()):
            found.append(file)

    return found



from typing import List

from guet.commands import CommandFactory
from guet.files import FileSystem
from guet.git import Git
from guet.steps import OptionStep, Step
from guet.steps.action import PrintAction
from guet.steps.check import GitRequiredCheck, HelpCheck, VersionCheck
from guet.steps.preparation import InitializePreparation
from guet.util import FlagBuilder, FlagsBuilder, HelpMessageBuilder

from ._cancel_create_hook import CancelCreateHooks
from ._change_hooks_folder import ChangeHooksFolder
from ._create_hooks_alongside import CreateHooksAlongside
from ._create_hooks_normally import CreateHooksNormally

ALONGSIDE_FLAG_EXPLANATION = 'Create hooks alongside current hooks with "-guet" on the end'
SHORT_EXPLANATION = 'Initialize current .git project to use guet.'
START_HELP_MESSAGE = HelpMessageBuilder('guet start', SHORT_EXPLANATION) \
    .flags(FlagsBuilder([FlagBuilder('-a/--alongside', ALONGSIDE_FLAG_EXPLANATION),
                         FlagBuilder('-o/--overwrite', 'Overwrite current hooks'),
                         FlagBuilder('--location', 'Specify location to create hooks in')])).build()


def choose(args: List[str]):
    if '-a' in args or '--alongside' in args:
        return 1
    elif '-o' in args or '--overwrite' in args:
        return 0

    else:
        return 0


class InitCommandFactory(CommandFactory):
    def __init__(self, git: Git, file_system: FileSystem):
        self.git = git
        self.file_system = file_system

    def choose(self, args: List[str]):
        res = 0
        if '-a' in args or '--alongside' in args:
            res = 1
        elif '-o' in args or '--overwrite' in args:
            res = 0
        elif self.git.non_guet_hooks_present():
            res = self._capture_overwrite()
        else:
            res = 0
        return res

    def _capture_overwrite(self):
        print('There is already commit hooks in this project. You can')
        print('  (o) overwrite current hooks. This will delete any'
              ' matching hooks.')
        print(('  (a) create guet hooks alongside current ones.'
               'This will create them with \'-guet\' appended on'
               ' the name of the hook file.'))
        print('  (x) cancel the request. This will do nothing.')
        val = input()
        if val == 'o':
            return 0
        elif val == 'a':
            return 1
        else:
            return 2

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(START_HELP_MESSAGE)) \
            .next(InitializePreparation(self.file_system)) \
            .next(GitRequiredCheck(self.git)) \
            .next(ChangeHooksFolder(self.git)) \
            .next(OptionStep(
                [
                    CreateHooksNormally(self.git)
                    .next(PrintAction('guet successfully started in this repository.')),
                    CreateHooksAlongside(self.git)
                    .next(PrintAction('guet successfully started in this repository.')),
                    CancelCreateHooks()
                ],
                self.choose
            ))



from os.path import join
from pathlib import Path

from guet import constants
from guet.config import CONFIGURATION_DIRECTORY
from guet.files.write_lines import write_lines
from guet.settings.settings import Settings


def set_settings(settings: Settings) -> None:
    lines = settings.write()
    write_lines(Path(join(CONFIGURATION_DIRECTORY, constants.CONFIG)), lines)



from guet.commands import CommandFactory, CommandMap
from guet.steps import Step
from guet.steps.action import PrintAction

from ._usage import UsageAction


class UnknownCommandFactory(CommandFactory):
    def __init__(self, command_map: CommandMap):
        self.command_map = command_map

    def build(self) -> Step:
        return PrintAction('Not a valid guet command.\n') \
            .next(UsageAction(self.command_map))



from typing import List

from guet.committers import Committers2 as Committers
from guet.steps.check import Check


class OverwritingCommitterCheck(Check):

    def __init__(self, committers: Committers):
        super().__init__('')
        self.committers = committers

    def should_stop(self, args: List[str]) -> bool:
        initials, name, email, *_ = args
        found = self.committers.by_initials(args[0])
        if not found:
            return False
        else:
            print((f'Matching initials "{initials}". Adding '
                   f'"{name}" <{email}> will overwrite '
                   f'"{found.name}" <{found.email}>. Would you '
                   'like to continue(y) or cancel(x)?'))

            choice = input()

            if choice == 'x':
                return True
            else:
                self.committers.remove(found.initials)
                return False



class Args(list):

    def __init__(self, raw_args):
        super().__init__()
        self._raw_args = raw_args

    def __setitem__(self, index, data):
        self._raw_args[index] = data

    def __getitem__(self, index):
        return self._raw_args[index]

    def __len__(self):
        return len(self._raw_args)

    @property
    def without_flags(self):
        new_args = []
        for arg in self._raw_args:
            if not arg.startswith('-'):
                new_args.append(arg)
        return new_args



from typing import List

from guet.commands import CommandMap
from guet.steps.action import Action


class UsageAction(Action):
    def __init__(self, command_map: CommandMap):
        super().__init__()
        self.command_map = command_map

    def execute(self, args: List[str]):
        print('usage: guet <command>\n')
        for key in self.command_map.all_commands():
            print(f'{key}: {self.command_map.get_description(key)}')



class Proxy:
    def __init__(self, proxy_object=None):
        self.proxy_object = proxy_object

    def __getattribute__(self, name):
        # This won't load self.proxy_object if it's a propery that's accessed.

        if name == 'proxy_object':
            return object.__getattribute__(self, 'proxy_object')
        if name == 'loader':
            return object.__getattribute__(self, 'loader')
        if self.proxy_object is None:
            self.proxy_object = self.loader()
        return getattr(self.proxy_object, name)

    def loader(self):
        raise NotImplementedError



from os import getcwd
from pathlib import Path
from typing import List

from guet.git import Git
from guet.steps.preparation import Preparation


class ChangeHooksFolder(Preparation):
    def __init__(self, git: Git):
        super().__init__()
        self._git = git

    def prepare(self, args: List[str]):
        if '--location' in args:
            location_flag_position = args.index('--location')
            if (location_flag_position + 1) >= len(args):
                print(('No argument following hook location flag. '
                       'Default hooks path used instead.'))
            else:
                self._change_hooks_directory(args)

    def _change_hooks_directory(self, args: List[str]):
        location_flag_position = args.index('--location')
        maybe_path = Path(getcwd(), args[location_flag_position + 1])
        if maybe_path.is_dir():
            self._git.set_hooks_destination(maybe_path)
        else:
            print((f'Not a folder: {maybe_path.absolute()}'
                   '. Default hooks path used instead.'))



from typing import List

from guet.commands import CommandFactory
from guet.committers import Committers2
from guet.files import FileSystem
from guet.git import Git
from guet.steps import IfStep, Step
from guet.steps.check import GitRequiredCheck, HelpCheck, VersionCheck
from guet.steps.preparation import InitializePreparation
from guet.util import FlagBuilder, FlagsBuilder, HelpMessageBuilder

from ._add_committer import AddCommitter
from ._args import ArgumentCheck
from ._local_file_initialization import LocalFilesInitialization
from ._overwrite import OverwritingCommitterCheck

_HELP_MESSAGE = HelpMessageBuilder('guet add <initials> <"name"> <email>',
                                   'Add committer to make available for commit tracking.') \
    .flags(FlagsBuilder([FlagBuilder('--local', 'Save the committer locally')])).build()

class AddCommandFactory(CommandFactory):
    def __init__(self, file_system: FileSystem, committers: Committers2, git: Git):
        self.file_system = file_system
        self.committers = committers
        self.git = git

    def choose(self, args: List[str]):
        return 1 if '--local' in args else 0

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(_HELP_MESSAGE)) \
            .next(InitializePreparation(self.file_system)) \
            .next(ArgumentCheck()) \
            .next(IfStep(lambda args: '--local' in args,
                         GitRequiredCheck(self.git)
                         .next(LocalFilesInitialization(self.file_system, self.committers)))) \
            .next(OverwritingCommitterCheck(self.committers)) \
            .next(AddCommitter(self.committers))



from pathlib import Path
from typing import List

from guet.committers import CurrentCommittersObserver
from guet.committers.committer import Committer
from guet.files.read_lines import read_lines
from guet.files.write_lines import write_lines
from guet.git._all_valid_hooks import all_valid_hooks, valid_hooks
from guet.git._author_manage import (append_new_author, get_author_lines,
                                     load_author, overwrite_current_author)
from guet.git._create_strategy import DoCreateStrategy, DontCreateStrategy
from guet.git._file_name_strategy import (AlongsideFileNameStrategy,
                                          BaseFileNameStrategy)
from guet.git._guet_hooks import GUET_HOOKS
from guet.git._hook_loader import HookLoader
from guet.git.author import Author
from guet.git.errors import NoGitPresentError
from guet.git.hook import Hook


def _load_hooks(hook_strategy: HookLoader) -> List[Hook]:
    hooks = []
    for hook_name in GUET_HOOKS:
        hook_strategy.apply(hook_name, hooks)
    return hooks


def _load_commit_msg(path_to_repository: Path) -> List[str]:
    try:
        return read_lines(path_to_repository.joinpath('COMMIT_EDITMSG'))
    except FileNotFoundError:
        return []


class Git(CurrentCommittersObserver):
    def __init__(self, repository_path: Path):
        super().__init__()
        if not repository_path.is_dir():
            raise NoGitPresentError()
        default = _load_hooks(HookLoader(
            repository_path.joinpath('hooks'), BaseFileNameStrategy(), DontCreateStrategy()))
        alongside = _load_hooks(HookLoader(
            repository_path.joinpath('hooks'), AlongsideFileNameStrategy(), DontCreateStrategy()))
        self.hooks = default + alongside
        self.path_to_git_folder = repository_path
        self._commit_msg = _load_commit_msg(repository_path)
        self._config_lines = read_lines(repository_path.joinpath('config'))
        self._author: Author = load_author(self._config_lines)
        self._hooks_destination = None

    @property
    def author(self) -> Author:
        return self._author

    @author.setter
    def author(self, new_author: Author):
        new_lines = list(self._config_lines)
        user, email = get_author_lines(self._config_lines)
        if user is not None and email is not None:
            overwrite_current_author(new_lines, new_author)
        else:
            append_new_author(new_lines, new_author)
        write_lines(self.path_to_git_folder.joinpath('config'), new_lines)
        self._author = new_author

    def hooks_destination(self):
        if self._hooks_destination:
            return self._hooks_destination
        else:
            return self.path_to_git_folder.joinpath('hooks')

    def set_hooks_destination(self, path: Path):
        self._hooks_destination = path

    @property
    def commit_msg(self):
        return self._commit_msg

    @commit_msg.setter
    def commit_msg(self, lines: List[str]):
        write_lines(self.path_to_git_folder.joinpath('COMMIT_EDITMSG'), lines)
        self._commit_msg = lines

    def hooks_present(self) -> bool:
        return all_valid_hooks(self.hooks)

    def non_guet_hooks_present(self) -> bool:
        non_guet_hook_found = False
        for hook in self.hooks:
            if not hook.is_guet_hook():
                non_guet_hook_found = True
        return non_guet_hook_found

    def create_hooks(self, alongside: bool = False) -> None:
        path = self.hooks_destination()
        if alongside:
            hook_loader = HookLoader(path, AlongsideFileNameStrategy(), DoCreateStrategy())
        else:
            hook_loader = HookLoader(path, BaseFileNameStrategy(), DoCreateStrategy())
        self.hooks = _load_hooks(hook_loader)
        for hook in self.hooks:
            hook.save()

    def guet_hooks(self) -> List[str]:
        return [hook.path for hook in valid_hooks(self.hooks)]

    def on_new_committers(self, committers: List[Committer]):
        author_committer = committers[0]
        self.author = Author(name=author_committer.name,
                             email=author_committer.email)



from typing import List

from guet.committers import Committers2 as Committers
from guet.committers import CommittersPrinter, CurrentCommitters
from guet.steps.action import Action


class GetCommittersAction(Action):

    def __init__(self,
                 committers: Committers,
                 current_committers: CurrentCommitters):
        super().__init__()
        self.committers = committers
        self.current = current_committers

    def execute(self, args: List[str]):
        printer = CommittersPrinter(initials_only=False)
        if args[0] == 'all':
            committers = self.committers.all()
            pre_print = 'All committers'
        else:
            committers = self.current.get()
            pre_print = 'Current committers'

        print(pre_print)
        printer.print(committers)



from os import chmod, stat
from pathlib import Path
from shutil import which
from typing import List

from guet.files.read_lines import read_lines
from guet.files.write_lines import write_lines


def shared_hook_lines() -> List[str]:
    return [
        'from guet.hooks import run',
        'import sys',
        'run(sys.argv[0])',
    ]


PYTHON_GUET_HOOK = ['#! /usr/bin/env python'] + shared_hook_lines()

PYTHON3_GUET_HOOK = ['#! /usr/bin/env python3'] + shared_hook_lines()


class Hook:

    def __init__(self, path_to_hook: Path, *, create: bool = False):
        self.path = path_to_hook
        self.content = self._parse_file_content(create, path_to_hook)

    def __repr__(self) -> str:
        return f'Hook: path: {self.path}, content: {self.content}'

    def is_guet_hook(self):
        return self.content == PYTHON3_GUET_HOOK or self.content == PYTHON_GUET_HOOK

    @staticmethod
    def _parse_file_content(create, path_to_hook):
        _content = Hook._get_file_content(path_to_hook, create)
        if _content not in (PYTHON3_GUET_HOOK, PYTHON_GUET_HOOK):
            _content = Hook._handle_mismatched_content(_content, create)
        return _content

    @staticmethod
    def _handle_mismatched_content(_content, create):
        if create:
            _content = PYTHON3_GUET_HOOK
        return _content

    @staticmethod
    def _get_file_content(path_to_hook, create: bool):
        try:
            _content = read_lines(path_to_hook)
        except FileNotFoundError:
            if create:
                if not which('python3'):
                    _content = PYTHON_GUET_HOOK
                else:
                    _content = PYTHON3_GUET_HOOK
            else:
                raise
        return _content

    def save(self):
        write_lines(self.path, self.content)
        status = stat(str(self.path))
        chmod(str(self.path), status.st_mode | 0o111)



from .action import Action


class PrintAction(Action):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def execute(self, _):
        print(self.message)



from typing import List

from guet import __version__ as version
from guet.steps.check.check import Check


class VersionCheck(Check):

    def __init__(self):
        super().__init__(version)

    def should_stop(self, args: List[str]) -> bool:
        return '--version' in args or '-v' in args



from typing import List

from guet.settings.boolean_parser import boolean_parser
from guet.settings.setting import Setting


class Settings:

    def __init__(self):
        self.version = None
        self._settings = dict()
        self._settings['pairReset'] = Setting(default_value=True,
                                              parser=boolean_parser,
                                              validator=lambda v: isinstance(v, bool))
        self._settings['debug'] = Setting(default_value=False,
                                          parser=boolean_parser,
                                          validator=lambda v: isinstance(v, bool))

    def load(self, configuration_file_lines: List[str]):
        lines_without_version = configuration_file_lines[2:]
        self.version = configuration_file_lines[0].rstrip()
        for attribute in lines_without_version:
            self._load_attribute(attribute)

    def _load_attribute(self, attribute):
        key, value = attribute.rstrip().split('=')
        try:
            self.set(key, value)
        except KeyError:
            print(f'Unknown configuration value \"{key}\" in configuration file.')
            exit(1)

    def write(self) -> List[str]:
        version = [f'{self.version}\n', '\n']
        settings = []
        for key in self._settings:
            if not self._settings[key].is_default_value():
                settings.append(self._convert_to_line(key))
        return version + settings

    def set(self, key: str, value) -> None:
        read_writer = self._settings[key]
        read_writer.value = read_writer.parser(value)
        self._settings[key] = read_writer

    def read(self, key: str) -> Setting:
        return self._settings[key].value

    def _convert_to_line(self, attribute_key: str) -> str:
        return f'{attribute_key}={self.read(attribute_key)}\n'



from typing import List


def get_command_key(args: List[str]) -> str:
    if len(_strip_flags(args)) == 0:
        return 'help'
    else:
        return args[0]


def _strip_flags(args: List[str]) -> List[str]:
    return [arg for arg in args if not arg.startswith('-')]



from guet.committers.committer import Committer
from guet.files import FileSystem

from ._global_committer_state import GlobalCommittersState
from ._local_committers_state import LocalCommittersState


class Committers:
    def __init__(self, file_system: FileSystem):
        self.global_state = GlobalCommittersState(file_system)
        self.local_state = LocalCommittersState(file_system, self.global_state)

        self.current_state = self.global_state

    def to_local(self):
        self.current_state = self.local_state

    def to_global(self):
        self.current_state = self.global_state

    def all(self):
        def sort_key(committer: Committer):
            return committer.initials
        found = self.current_state.all()
        found.sort(key=sort_key)
        return found

    def by_initials(self, initials: str):
        return self.current_state.by_initials(initials.lower())

    def add(self, committer: Committer):
        self.current_state.add(committer)

    def remove(self, initials: str):
        self.current_state.remove(initials)



from pathlib import Path
from shutil import rmtree
from typing import List

from guet.config import CONFIGURATION_DIRECTORY
from guet.steps.action import Action


class RemoveGlobal(Action):
    def execute(self, args: List[str]):
        if any(arg in ('-g', '--global') for arg in args):
            configuration_dir = Path(CONFIGURATION_DIRECTORY)
            rmtree(configuration_dir)
            print('Bye!')



from collections import namedtuple
from typing import List

from guet.commands import CommandFactory

_MappedCommand = namedtuple('MappedCommand', ['command', 'description'])


class CommandMap:
    def __init__(self):
        self._commands = dict()
        self._default = None

    def add_command(self, key: str, command: CommandFactory, short_description: str):
        self._commands[key] = _MappedCommand(command, short_description)

    def _get(self, key: str) -> _MappedCommand:
        return self._commands.get(key, self._default)

    def get_command(self, key: str) -> CommandFactory:
        return self._get(key).command

    def get_description(self, key: str) -> str:
        return self._get(key).description

    def all_commands(self) -> List[str]:
        return self._commands.keys()

    def set_default(self, command: CommandFactory):
        self._default = _MappedCommand(command, '')



from pathlib import Path
from typing import List


def read_lines(path: Path) -> List[str]:
    return _read_lines_from_path(path)


def _read_lines_from_path(path: Path) -> List[str]:
    all_lines = path.read_text().split('\n')
    if all_lines[len(all_lines) - 1] == '':
        del all_lines[len(all_lines) - 1]
    return all_lines



from typing import List

from guet.constants import COMMITTERS
from guet.files import File, FileSystem

from ._committers2_state import Committers2State
from .committer import Committer


class GlobalCommittersState(Committers2State):
    def __init__(self, file_system: FileSystem):
        super().__init__()
        self.file_system = file_system

    def all(self) -> List[Committer]:
        return [self._map_line_to_committer(c) for c in self._get_committers_file().read()]

    def add(self, committer: Committer):
        if self.by_initials(committer.initials):
            self.remove(committer.initials)
        all_committers = self.all()
        all_committers.append(committer)
        lines = [str(c) for c in all_committers]
        self._get_committers_file().write(lines)
        self._get_committers_file().save()

    def remove(self, initials: str):
        all_committers = self.all()
        without_given = [c for c in all_committers if c.initials != initials]
        lines = [str(c) for c in without_given]
        self._get_committers_file().write(lines)
        self._get_committers_file().save()

    def _get_committers_file(self) -> File:
        comitters_file_patah = FileSystem.config_directory().joinpath(COMMITTERS)
        return self.file_system.get(comitters_file_patah)



from os import mkdir
from os.path import isdir
from pathlib import Path
from typing import List

from guet.committers import Committers2 as Committers
from guet.files import FileSystem
from guet.steps.preparation import Preparation
from guet.util import project_root


class LocalFilesInitialization(Preparation):
    def __init__(self, file_system: FileSystem, committers: Committers):
        super().__init__()
        self._file_system = file_system
        self.committers = committers

    def prepare(self, args: List[str]):
        if '--local' in args:
            self.committers.to_local()
            config_dir = Path(project_root()).joinpath('.guet')
            if not isdir(config_dir):
                mkdir(config_dir)
                self._create_empty_file(config_dir.joinpath('committers'))
                self._file_system.save_all()

    def _create_empty_file(self, path: Path) -> None:
        file = self._file_system.get(path)
        file.read()
        return file



from typing import List

from guet.steps.action import Action


class RemoveCommitterAction(Action):
    def __init__(self, committers):
        super().__init__()
        self.committers = committers

    def execute(self, args: List[str]):
        committer = self.committers.by_initials(args[0])
        if not committer:
            print(f'No committer exists with initials {args[0]}')
        else:
            self.committers.remove(committer.initials)



from typing import List

from guet.committers import Committers2 as Committers
from guet.committers import CommittersPrinter, CurrentCommitters
from guet.steps.action import Action


class SetCommittersAction(Action):
    def __init__(self,
                 committers: Committers,
                 current_committers: CurrentCommitters):
        super().__init__()
        self.committers = committers
        self.current_committers = current_committers

    def execute(self, args: List[str]):
        lowercase_args = [arg.lower() for arg in args]
        found = [c for c in self.committers.all() if c.initials in lowercase_args]
        self.current_committers.set(found)
        printer = CommittersPrinter(initials_only=False)
        print('Committers set to:')
        printer.print(found)



from pathlib import Path
from typing import List

from guet.git._create_strategy import CreateStrategy
from guet.git._file_name_strategy import FileNameStrategy
from guet.git.errors import NotGuetHookError
from guet.git.hook import Hook


class HookLoader:
    def __init__(self,
                 hooks_dir: Path,
                 file_name: FileNameStrategy,
                 create: CreateStrategy):
        self.create = create
        self.file_name = file_name
        self._hooks_directory: Path = hooks_dir

    def _create_hooks_directory_if_not_present(self):
        if not self._hooks_directory.is_dir():
            self._hooks_directory.mkdir()
            print('No hooks directory found, creating one.')

    def apply(self, hook_name: str, hooks: List[Hook]):
        self._create_hooks_directory_if_not_present()
        hook_path = self._hooks_directory.joinpath(self.file_name.apply(hook_name))
        try:
            hooks.append(Hook(hook_path, create=self.create.apply()))
        except FileNotFoundError:
            pass
        except NotGuetHookError:
            pass



from guet.commands import CommandFactory, CommandMap
from guet.files import FileSystem
from guet.steps import Step
from guet.steps.check import VersionCheck
from guet.steps.preparation import InitializePreparation

from ._usage import UsageAction


class HelpCommandFactory(CommandFactory):
    def __init__(self, command_map: CommandMap, file_system: FileSystem):
        self.command_map = command_map
        self.file_system = file_system

    def build(self) -> Step:
        return VersionCheck() \
            .next(InitializePreparation(self.file_system)) \
            .next(UsageAction(self.command_map))



from typing import Callable, List

from guet.steps import Step


class IfStep(Step):
    def __init__(self, if_true: Callable[[List[str]], bool], steps: Step):
        super().__init__()
        self.if_true = if_true
        self.steps = steps

    def do_play(self, args: List[str]):
        if self.if_true(args):
            self.steps.play(args)



from typing import List

from guet.steps.step import Step


class Preparation(Step):

    def prepare(self, args: List[str]):
        raise NotImplementedError()

    def do_play(self, args: List[str]):
        self.prepare(args)



from ._committer_print_formatter import MultipleCommitterPrintFormatter


class InitialsFormatter(MultipleCommitterPrintFormatter):
    def __str__(self):
        return ', '.join([committer.initials for committer in self._committers])



from typing import List

from guet.steps.check.check import Check


class HelpCheck(Check):

    def __init__(self, stop_message, *, stop_on_no_args=False):
        super().__init__(stop_message)
        self.stop_on_no_args = stop_on_no_args

    def should_stop(self, args: List[str]) -> bool:
        return '--help' in args or '-h' in args or self._should_stop_for_empty_args(args)

    def _should_stop_for_empty_args(self, args: List[str]) -> bool:
        if self.stop_on_no_args:
            return len(args) == 0
        return False



from guet.git.git import Git
from guet.steps.check.check import Check

GIT_REQUIRED_MESSAGE = "Git not installed in this directory."


class GitRequiredCheck(Check):

    def __init__(self, git: Git):
        super().__init__(GIT_REQUIRED_MESSAGE)
        self.git = git

    def should_stop(self, args):
        try:
            # This depends on GitProxy loading it's Git internally, causing the FileNotFoundError.
            self.git.hooks_present()
            return False
        except FileNotFoundError:
            return True



from typing import List

from guet.committers.committer import Committer

from ._initials_formatter import InitialsFormatter
from ._initials_name_email_printer import InitialsNameEmailPrintFormatter


class CommittersPrinter:
    def __init__(self, *, initials_only: bool):
        self.initials_only = initials_only

    def print(self, committers: List[Committer]):
        if self.initials_only:
            self._initials_print(committers)
        else:
            self._pretty_print(committers)

    def _initials_print(self, committers: List[Committer]):
        print(str(InitialsFormatter(committers)))

    def _pretty_print(self, committers: List[Committer]):
        for committer in committers:
            print(str(InitialsNameEmailPrintFormatter(committer)))



from typing import List

from guet.committers import CurrentCommitters
from guet.steps.action import Action


class PreCommit(Action):
    def __init__(self, current: CurrentCommitters):
        super().__init__()
        self.current = current

    def execute(self, args: List[str]):
        if len(self.current.get()) == 0:
            print('You must set your pairs before you can commit.')
            exit(1)



from guet.commands import CommandFactory
from guet.steps import Step
from guet.steps.check import (CommittersExistCheck, GitRequiredCheck,
                              HelpCheck, VersionCheck)
from guet.steps.preparation import InitializePreparation, SwapToLocal
from guet.util import HelpMessageBuilder

from ._set_committers import SetCommittersAction

SET_HELP_MESSAGE = HelpMessageBuilder(
    'guet set <initials> [<initials> ...]', 'Set current committers.').build()


class SetCommittersCommand(CommandFactory):
    def __init__(self, file_system, committers, current, git):
        self.committers = committers
        self.current = current
        self.file_system = file_system
        self.git = git

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(SET_HELP_MESSAGE, stop_on_no_args=True)) \
            .next(InitializePreparation(self.file_system)) \
            .next(GitRequiredCheck(self.git)) \
            .next(SwapToLocal(self.committers)) \
            .next(CommittersExistCheck(self.committers)) \
            .next(SetCommittersAction(self.committers, self.current))



import time
from os.path import join
from pathlib import Path
from typing import List

from guet import constants
from guet.committers._committers_set import CommittersSet, all_committers_set
from guet.committers.committer import Committer
from guet.config import CONFIGURATION_DIRECTORY
from guet.files.write_lines import write_lines


def _format_committers_to_committers_set_format(committers: List[Committer],
                                                project_path: Path) -> CommittersSet:
    git_path = project_path
    current_time_in_millis = int(round(time.time() * 1000))
    committer_initials = [committer.initials for committer in committers]
    return CommittersSet(committer_initials, current_time_in_millis, git_path)


def _index_with_matching_path(current_set: List[CommittersSet],
                              formatted_set_committers_information: CommittersSet):
    for index, current in enumerate(current_set):
        if current.path == formatted_set_committers_information.path:
            return index
    return -1


def _add_to_current_set_lines(current_set: List[CommittersSet],
                              new_set: CommittersSet):
    index = _index_with_matching_path(current_set, new_set)
    if index != -1:
        _replace_committers_set_with_matching_initials(current_set, new_set, index)
    else:
        current_set.append(new_set)


def _replace_committers_set_with_matching_initials(current_committers_set: List[CommittersSet],
                                                   new_committers_set: CommittersSet,
                                                   index_of_match):
    current_committers_set[index_of_match] = new_committers_set


def _convert_to_text(current_set: CommittersSet) -> str:
    path = current_set.path
    current_time_in_millis = int(round(time.time() * 1000))
    committer_initials = ','.join(current_set.initials)
    return committer_initials + f',{current_time_in_millis},{path}'


def set_current_committers(committers: List[Committer], project_path: Path) -> None:
    path = Path(join(CONFIGURATION_DIRECTORY, constants.COMMITTERS_SET))
    current_set = all_committers_set()
    new_committers_set = _format_committers_to_committers_set_format(committers, project_path)
    _add_to_current_set_lines(current_set, new_committers_set)
    write_lines(path, [_convert_to_text(committers_set) for committers_set in current_set])



from typing import List

from guet.steps.check import Check
from guet.util import Args


class ArgumentCheck(Check):

    def __init__(self):
        super().__init__()

    def should_stop(self, args: List[str]) -> bool:
        return self.not_enough_args(args) or self.too_many_args(args)

    def load_message(self, args: List[str]) -> str:
        if self.not_enough_args(args):
            return 'Not enough arguments.'
        else:
            return 'Too many arguments.'

    def not_enough_args(self, args: List[str]) -> bool:
        return len(Args(args).without_flags) < 3

    def too_many_args(self, args: List[str]) -> bool:
        return len(Args(args).without_flags) > 3



from typing import List


class Step:

    def __init__(self):
        self._next = None

    def do_play(self, args: List[str]):
        raise NotImplementedError()

    def play(self, args: List[str]):
        self.do_play(args)
        if self._next is not None:
            self._next.play(args)

    def next(self, next_step: "Step") -> "Step":
        if self._next is None:
            self._next = next_step
        else:
            self._next.next(next_step)
        return self



from os import mkdir
from os.path import isdir
from pathlib import Path
from typing import List

from guet.steps.preparation.preapration import Preparation


class LocalPreparation(Preparation):
    def __init__(self, project_root: Path):
        super().__init__()
        self._project_root = project_root

    def prepare(self, args: List[str]):
        local_guet_directory = self._project_root.joinpath('.guet')
        if not isdir(local_guet_directory):
            mkdir(local_guet_directory)



from os.path import isdir
from typing import List

from guet.committers import Committers2 as Committers
from guet.steps.preparation.preapration import Preparation
from guet.util import project_root


class SwapToLocal(Preparation):
    def __init__(self, committers: Committers):
        super().__init__()
        self.committers = committers

    def prepare(self, args: List[str]):
        try:
            local_guet_directory = project_root().joinpath('.guet')
            if isdir(local_guet_directory):
                self.committers.to_local()
        except FileNotFoundError:
            pass



from guet.commands import CommandFactory
from guet.steps import Step
from guet.steps.check import CommittersExistCheck, HelpCheck, VersionCheck
from guet.steps.preparation import InitializePreparation
from guet.util import HelpMessageBuilder

from ._remove_committer import RemoveCommitterAction

REMOVE_HELP_MESSAGE = HelpMessageBuilder('guet remove <initials>',
                                         ('Remove committer '
                                          'with given initials from system.')).build()


class RemoveCommandFactory(CommandFactory):
    def __init__(self, file_system, committers):
        self.committers = committers
        self.file_system = file_system

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(REMOVE_HELP_MESSAGE, stop_on_no_args=True)) \
            .next(InitializePreparation(self.file_system)) \
            .next(CommittersExistCheck(self.committers)) \
            .next(RemoveCommitterAction(self.committers))



from typing import List

from guet.constants import COMMITTERS
from guet.files import File, FileSystem
from guet.util import project_root

from ._committers2_state import Committers2State
from ._global_committer_state import GlobalCommittersState
from .committer import Committer


class LocalCommittersState(Committers2State):
    def __init__(self, file_system: FileSystem, global_state: GlobalCommittersState):
        super().__init__()
        self.file_system = file_system
        self.global_state = global_state

    def all(self) -> List[Committer]:
        local_committers = self._all_local_committers()
        global_committers = self._remove_globals_with_initials_in_local(
            self.global_state.all(), local_committers)

        return global_committers + local_committers

    def add(self, committer: Committer):
        if self.global_state.by_initials(committer.initials):
            print(f'Adding committer with initials "{committer.initials}" '
                  'will overshadow global committer with same initials.')
        if self._local_committer_present(committer.initials):
            self.remove(committer.initials)
        all_committers = self._all_local_committers()
        all_committers.append(committer)
        lines = [str(c) for c in all_committers]
        self._get_committers_file().write(lines)
        self._get_committers_file().save()

    def remove(self, initials: str):
        all_committers = self._all_local_committers()
        without_given = [c for c in all_committers if c.initials != initials]
        lines = [str(c) for c in without_given]
        self._get_committers_file().write(lines)
        self._get_committers_file().save()

    def _get_committers_file(self) -> File:
        pass

    def _all_local_committers(self):
        return [self._map_line_to_committer(c) for c in self._get_committers_file().read()]

    def _remove_globals_with_initials_in_local(self, global_committers, local_committers):
        local_initials = [committer.initials for committer in local_committers]
        return [c for c in global_committers if c.initials not in local_initials]

    def _local_committer_present(self, initials) -> bool:
        try:
            return next((c for c in self._all_local_committers() if c.initials == initials))
        except StopIteration:
            return None

    def _get_committers_file(self) -> File:
        comitters_file_patah = project_root().joinpath('.guet').joinpath(COMMITTERS)
        return self.file_system.get(comitters_file_patah)



from typing import List

from guet.committers import CurrentCommitters
from guet.steps.action import Action


class PostCommit(Action):
    def __init__(self, current_committers: CurrentCommitters):
        super().__init__()
        self.current_committers = current_committers

    def execute(self, args: List[str]):
        found = self.current_committers.get()
        if len(found) > 1:
            first = found.pop(0)
            found.append(first)
            self.current_committers.set(found)



from pathlib import Path

from guet.config import CONFIGURATION_DIRECTORY
from guet.files import File


class _FileMap(dict):
    def __missing__(self, key: Path):
        self[key] = File(key)
        return self[key]


class FileSystem:
    def __init__(self):
        self._file_cache: _FileMap = _FileMap()

    def get(self, path: Path) -> File:
        return self._file_cache[path]

    def save_all(self):
        for path in self._file_cache:
            file = self._file_cache[path]
            file.save()

    @staticmethod
    def config_directory() -> Path:
        return Path(CONFIGURATION_DIRECTORY)



from guet.committers._comitters_printer import CommittersPrinter
from guet.committers._committer_print_formatter import CommitterPrintFormatter
from guet.committers._initials_formatter import InitialsFormatter
from guet.committers._initials_name_email_printer import \
    InitialsNameEmailPrintFormatter

from ._committers2 import Committers as Committers2
from ._current_committers import CurrentCommitters, CurrentCommittersObserver



from typing import List

from guet.steps.step import Step


class Check(Step):
    def __init__(self, stop_message: str = None):
        super().__init__()
        self._stop_message = stop_message

    def do_play(self, args: List[str]):
        if self.should_stop(args):
            message = self._stop_message
            if message is None:
                message = self.load_message(args)
            print(message)
            exit(1)

    def should_stop(self, args: List[str]) -> bool:
        raise NotImplementedError()

    def load_message(self, _: List[str]) -> str:
        return ""



from typing import List


def add_command_help_if_invalid_command_given(args: List[str]) -> List[str]:
    if len(args) == 0:
        return ['help']
    if args[0].startswith('-'):
        return ['help'] + args
    else:
        return args



from os import remove
from pathlib import Path
from typing import Callable, List

from guet.files.read_lines import read_lines
from guet.files.write_lines import write_lines


class File:
    def __init__(self, absolute_path: Path):
        self._content = None
        self._absolute_path = absolute_path
        self._marked_for_deletion = False
        self._written_to = False

    def read(self) -> List[str]:
        try:
            self._content = read_lines(self._absolute_path)
        except FileNotFoundError:
            self._content = []
        return self._content

    def write(self, new_content: List[str]):
        self._written_to = True
        self._content = new_content
        self._marked_for_deletion = False

    def save(self):
        if self._marked_for_deletion:
            remove(self._absolute_path)
        elif self._content is not None and self._written_to:
            write_lines(self._absolute_path, self._content)

    def overwrite(self, matcher: Callable[[str], bool], new_line: str):
        for index, line in enumerate(self.read()):
            if matcher(line):
                self._content[index] = new_line

    def delete(self):
        self._marked_for_deletion = True



from typing import List

from .committer import Committer


class Committers2State:
    def all(self) -> List[Committer]:
        raise NotImplementedError

    def by_initials(self, initials: str) -> Committer:
        try:
            return next(c for c in self.all() if c.initials == initials)
        except StopIteration:
            return None

    def add(self, committer: Committer):
        raise NotImplementedError

    def remove(self, initials: str):
        raise NotImplementedError

    def _map_line_to_committer(self, line: str) -> Committer:
        initials, name, email = line.split(',')
        return Committer(initials=initials, name=name, email=email)



from typing import List

from guet.committers import Committers2 as Committers

from .check import Check


class CommittersExistCheck(Check):
    def __init__(self, committers: Committers):
        super().__init__()
        self.committers = committers

    def should_stop(self, args: List[str]) -> bool:
        return len(self._get_missing_committer_initials(args)) > 0

    def load_message(self, args: List[str]) -> str:
        return '\n'.join([f"No committer exists with initials '{initial}'"
                          for initial in
                          self._get_missing_committer_initials(args)])

    def _get_missing_committer_initials(self, args: List[str]) -> List[str]:
        missing = []
        for initial in args:
            if not self.committers.by_initials(initial):
                missing.append(initial)
        return missing



from os.path import isdir
from shutil import rmtree
from typing import List

from guet.files import FileSystem
from guet.git import Git, all_guet_hooks
from guet.steps.action import Action
from guet.util import project_root


class RemoveLocal(Action):
    def __init__(self, git: Git, file_system: FileSystem):
        super().__init__()
        self.git = git
        self.file_system = file_system

    def execute(self, args: List[str]):
        for hook in all_guet_hooks(self.file_system):
            hook.delete()
        if isdir(project_root().joinpath('.guet')):
            rmtree(project_root().joinpath('.guet'))
            print('guet tracking removed from this repository')
        else:
            print('No local guet configurations for this project')



from pathlib import Path
from typing import List


def _append_newline_if_not_present(line: str):
    if line.endswith('\n'):
        return line
    else:
        return f'{line}\n'


def _format_lines(lines: List[str]) -> str:
    return ''.join([_append_newline_if_not_present(line) for line in lines])


def _write_lines_to_path(path: Path, lines: List[str]) -> None:
    text = _format_lines(lines)
    path.write_text(text)


def write_lines(path: Path, lines: List[str]) -> None:
    return _write_lines_to_path(path, lines)



from guet.committers._committer_print_formatter import \
    SingleCommitterPrintFormatter


class InitialsNameEmailPrintFormatter(SingleCommitterPrintFormatter):
    def __str__(self):
        return f'{self._committer.initials} - {self._committer.name} <{self._committer.email}>'



from os.path import expanduser
from pathlib import Path


def recursive_directory_find(starting_path: Path, directory_name: str) -> Path:
    if str(starting_path) == expanduser('~'):
        raise FileNotFoundError()
    joined = starting_path.joinpath(directory_name)
    if joined.is_dir():
        return starting_path
    else:
        return recursive_directory_find(starting_path.parent, directory_name)



from typing import Union

from guet.committers import Committers2 as Committers
from guet.committers import CurrentCommitters
from guet.files import FileSystem
from guet.git import Git
from guet.steps import Step
from guet.util import project_root

from ._commit_msg import CommitMsg
from ._post_commit import PostCommit
from ._pre_commit import PreCommit

FILE_SYSTEM = FileSystem()
COMMITTERS = Committers(FILE_SYSTEM)
GIT = Git(project_root().joinpath('.git'))
CURRENT_COMMITTERS = CurrentCommitters(FILE_SYSTEM, COMMITTERS)
CURRENT_COMMITTERS.register_observer(GIT)


def run(hook: str):
    command = _choose_command(hook)
    if command:
        command.play([])
    FILE_SYSTEM.save_all()


def _choose_command(hook: str) -> Union[Step, None]:
    if hook.endswith('pre-commit'):
        return PreCommit(CURRENT_COMMITTERS)
    elif hook.endswith('post-commit'):
        return PostCommit(CURRENT_COMMITTERS)
    elif hook.endswith('commit-msg'):
        return CommitMsg(CURRENT_COMMITTERS, GIT)
    return None



from typing import List

from guet.committers import Committers2 as Committers
from guet.committers.committer import Committer
from guet.files import FileSystem
from guet.util import project_root

from ._initials_for_project import initials_for_project
from ._set_current_committers import set_current_committers


class CurrentCommittersObserver:
    def on_new_committers(self, committers: List[Committer]):
        pass


class CurrentCommitters:
    def __init__(self, file_system: FileSystem, committers: Committers):
        self.committers = committers
        self.file_system = file_system
        self.observers = []

    def get(self) -> List[Committer]:
        project_initials = initials_for_project(project_root())
        committers = []
        for initials in project_initials:
            committers.append(self.committers.by_initials(initials))
        return committers

    def set(self, committers: List[Committer]):
        set_current_committers(committers, project_root())
        for observer in self.observers:
            observer.on_new_committers(committers)

    def register_observer(self, observer: CurrentCommittersObserver):
        self.observers.append(observer)



from typing import List

from guet.git import Git
from guet.steps.action import Action


class CreateHooksAlongside(Action):
    def __init__(self, git: Git):
        super().__init__()
        self.git = git

    def execute(self, args: List[str]):
        self.git.create_hooks(alongside=True)



from os.path import join
from pathlib import Path
from typing import List, NamedTuple

from guet import constants
from guet.config import CONFIGURATION_DIRECTORY
from guet.files.read_lines import read_lines

CommittersSet = NamedTuple('CommittersSet',
                           [('initials', List[str]), ('set_time', int), ('path', Path)])


def all_committers_set() -> List[CommittersSet]:
    lines = read_lines(Path(join(CONFIGURATION_DIRECTORY, constants.COMMITTERS_SET)))
    committers_sets = []
    for line in lines:
        *initials, set_time, path_as_str = line.split(',')
        committers_sets.append(CommittersSet(initials, set_time, Path(path_as_str)))
    return committers_sets



from typing import List

from guet.committers import Committers2 as Committers
from guet.committers.committer import Committer
from guet.steps.action import Action
from guet.util import Args


class AddCommitter(Action):

    def __init__(self, committers: Committers):
        super().__init__()
        self.committers = committers

    def execute(self, args: List[str]):
        initials, name, email = Args(args).without_flags
        self.committers.add(Committer(name, email, initials))



from typing import List

from guet.commands import CommandFactory
from guet.files import FileSystem
from guet.git import Git
from guet.steps import Step
from guet.steps.check import GitRequiredCheck, HelpCheck, VersionCheck
from guet.steps.preparation import InitializePreparation
from guet.util import FlagBuilder, FlagsBuilder, HelpMessageBuilder

from ._remove_global import RemoveGlobal
from ._remove_local import RemoveLocal

_GLOBAL_EXPLANATION = 'Remove guet configuration from home directory'

_HELP_MESSAGE = HelpMessageBuilder('guet yeet',
                                   'Remove guet configuration.') \
    .flags(FlagsBuilder([FlagBuilder('-g/--global', _GLOBAL_EXPLANATION)])).build()


class YeetCommandFactory(CommandFactory):
    def __init__(self, file_system: FileSystem, git: Git):
        self.file_system = file_system
        self.git = git

    def choose(self, args: List[str]):
        return 1 if '--local' in args else 0

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(_HELP_MESSAGE)) \
            .next(InitializePreparation(self.file_system)) \
            .next(GitRequiredCheck(self.git)) \
            .next(RemoveLocal(self.git, self.file_system)) \
            .next(RemoveGlobal())



from setuptools import setup

import guet

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setup(name='guet',
      version=guet.__version__,
      description='Enable contribution tracking when pair programming',
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      url='https://github.com/chiptopher/guet',
      keywords='pair programming',
      entry_points={'console_scripts': ['guet=guet.main:main']},
      packages=[
          'guet',
          'guet.commands',
          'guet.commands.add',
          'guet.commands.get',
          'guet.commands.help',
          'guet.commands.init',
          'guet.commands.remove',
          'guet.commands.set',
          'guet.commands.yeet',
          'guet.committers',
          'guet.config',
          'guet.files',
          'guet.git',
          'guet.hooks',
          'guet.settings',
          'guet.steps',
          'guet.steps.action',
          'guet.steps.check',
          'guet.steps.preparation',
          'guet.util'
      ])



from pathlib import Path
from typing import List

from guet.util import current_millis

from ._committers_set import all_committers_set

_TWENTY_FOUR_HOURS_IN_MILLISECONDS = 86400000


def _new_set_required(current_project_set) -> bool:
    reset_time = int(current_project_set.set_time) + _TWENTY_FOUR_HOURS_IN_MILLISECONDS
    return reset_time < current_millis()


def initials_for_project(project_root: Path) -> List[str]:
    try:
        current_project_set = next(
            c for c in all_committers_set() if c.path == project_root)

        if _new_set_required(current_project_set):
            return []
        else:
            return current_project_set.initials

    except StopIteration:
        return []



from typing import List

from guet.steps.step import Step


class Action(Step):

    def do_play(self, args: List[str]):
        self.execute(args)

    def execute(self, args: List[str]):
        raise NotImplementedError()



from typing import Callable, List

from guet.errors import UnexpectedError

from .step import Step


class OptionStep(Step):

    def __init__(self, choices: List[Step], choice: Callable[[List[str]], int]):
        super().__init__()
        self.choices = choices
        self.choice = choice

    def do_play(self, args: List[str]):
        desired_choice = self.choice(args)
        if self._has_desired_choice(desired_choice):
            self.choices[desired_choice].play(args)
        else:
            raise UnexpectedError(
                f'Attempting to choose {desired_choice} from {args}')

    def _has_desired_choice(self, position: int) -> bool:
        return (len(self.choices) - 1) >= position



from typing import List

from guet.committers import CurrentCommitters
from guet.committers.committer import Committer
from guet.git import Git
from guet.steps.action import Action


class CommitMsg(Action):
    def __init__(self, current_committers: CurrentCommitters, git: Git):
        super().__init__()
        self.current_committers = current_committers
        self.git = git

    def execute(self, _):
        lines = self._co_autor_lines(self.current_committers.get())
        self.git.commit_msg = lines

    def _co_autor_lines(self, committers: List[Committer]) -> List[str]:
        return [f'Co-authored-by: {committer.name} <{committer.email}>' for committer in committers]



from ._committers_exist import CommittersExistCheck
from .check import Check
from .git_required_check import GitRequiredCheck
from .help_check import HelpCheck
from .start_required_check import StartRequiredCheck
from .version_check import VersionCheck



from typing import List


class Committer:
    def __init__(self, name: str, email: str, initials: str):
        self.name = name
        self.email = email
        self.initials = initials.lower()

    def __eq__(self, o):
        return self.name == o.name and self.email == o.email and self.initials == o.initials

    def __str__(self):
        return f'{self.initials},{self.name},{self.email}'

    def __repr__(self):
        return f'<{str(self)}>'

    def save(self):
        raise NotImplementedError()


def filter_committers_with_initials(committers: List[Committer],
                                    initials: List[str]) -> List[Committer]:
    return [committer for committer in committers if committer.initials in initials]



from typing import List

from guet.git.git import Git
from guet.steps.check.check import Check
from guet.util import project_root

GUET_NOT_STARTED_ERROR = (
    'guet not initialized in this repository. '
    'Please use guet start to initialize repository '
    'for use with guet.'
)


class StartRequiredCheck(Check):
    def __init__(self):
        super().__init__(GUET_NOT_STARTED_ERROR)

    def should_stop(self, args: List[str]) -> bool:
        git = Git(project_root().joinpath('.git'))
        return not git.hooks_present()



from os.path import join
from pathlib import Path
from typing import List

from guet import constants
from guet.config import CONFIGURATION_DIRECTORY
from guet.files.write_lines import write_lines


def set_errors(error_lines: List[str]):
    write_lines(Path(join(CONFIGURATION_DIRECTORY, constants.ERRORS)), error_lines)



from os import getcwd
from pathlib import Path

from ._recursive_directory_find import recursive_directory_find


def project_root() -> Path:
    return recursive_directory_find(Path(getcwd()), '.git')



from guet.commands import CommandFactory
from guet.committers import Committers2 as Committers
from guet.committers import CurrentCommitters
from guet.files import FileSystem
from guet.steps import Step
from guet.steps.check import HelpCheck, VersionCheck
from guet.steps.preparation import InitializePreparation, SwapToLocal
from guet.util import HelpMessageBuilder

from ._action import GetCommittersAction

GET_HELP_MESSAGE = HelpMessageBuilder('guet get <identifier> [-flag, ...]',
                                      'Get currently set information.') \
    .explanation(('Valid Identifier\n\tcurrent - lists currently set committers'
                  '\n\tall - lists all committers')) \
    .build()


class GetCommandFactory(CommandFactory):
    def __init__(self,
                 file_system: FileSystem,
                 committers: Committers,
                 current_committers: CurrentCommitters):

        self.file_system = file_system
        self.committers = committers
        self.current = current_committers

    def build(self) -> Step:
        return VersionCheck() \
            .next(HelpCheck(GET_HELP_MESSAGE, stop_on_no_args=True)) \
            .next(InitializePreparation(self.file_system)) \
            .next(SwapToLocal(self.committers)) \
            .next(GetCommittersAction(self.committers, self.current))



import sys

from guet.commands import CommandMap
from guet.commands.add import AddCommandFactory
from guet.commands.get import GetCommandFactory
from guet.commands.help import HelpCommandFactory, UnknownCommandFactory
from guet.commands.init import InitCommandFactory
from guet.commands.remove import RemoveCommandFactory
from guet.commands.set import SetCommittersCommand
from guet.commands.yeet import YeetCommandFactory
from guet.committers import Committers2, CurrentCommitters
from guet.files import FileSystem
from guet.git import GitProxy
from guet.util import add_command_help_if_invalid_command_given
from guet.util.errors import log_on_error


@log_on_error
def main():
    file_system = FileSystem()
    committers = Committers2(file_system)
    git = GitProxy()
    current_committers = CurrentCommitters(file_system, committers)
    current_committers.register_observer(git)

    command_map = CommandMap()
    command_map.add_command('help', HelpCommandFactory(
        command_map, file_system), 'Display guet usage')
    command_map.add_command('init', InitCommandFactory(
        GitProxy(), file_system), 'Start guet tracking in the current repository')
    command_map.add_command('add', AddCommandFactory(
        file_system, committers, git), 'Add committer for tracking')
    command_map.add_command('get', GetCommandFactory(
        file_system, committers, current_committers), 'List information about committers')
    command_map.add_command('set', SetCommittersCommand(
        file_system, committers, current_committers, git), 'Set committers for current repository')
    command_map.add_command('remove', RemoveCommandFactory(
        file_system, committers), 'Remove committer')
    command_map.add_command('yeet',
                            YeetCommandFactory(file_system, git),
                            'Remove guet configurations')

    command_map.set_default(UnknownCommandFactory(command_map))

    args = add_command_help_if_invalid_command_given(sys.argv[1:])

    command = command_map.get_command(args[0]).build()
    command.play(args[1:])

    file_system.save_all()



from os.path import join
from pathlib import Path
from typing import List

from guet import constants
from guet.config import CONFIGURATION_DIRECTORY
from guet.files.read_lines import read_lines
from guet.files.write_lines import write_lines

_COMMITTER_NOT_PRESENT = -1

_GLOBAL = Path(join(CONFIGURATION_DIRECTORY, constants.COMMITTERS))


def add_committer(initials: str, name: str, email: str, *, file_path: Path = _GLOBAL) -> None:
    all_committers = _read_all_committers_from_file(file_path)
    _add_committer_to_committers(all_committers, initials, name, email)
    write_lines(file_path, all_committers)


def _add_committer_to_committers(all_committers: List[str], initials: str, name: str, email: str):
    committer_formatted = f'{initials},{name},{email}\n'
    committer_position = _position_of_committer_with_initials(all_committers, initials)
    if committer_position is _COMMITTER_NOT_PRESENT:
        all_committers.append(committer_formatted)
    else:
        all_committers[committer_position] = committer_formatted


def _position_of_committer_with_initials(all_committers: List[str], initials: str) -> int:
    for index, committer in enumerate(all_committers):
        if committer.startswith(initials):
            return index
    return _COMMITTER_NOT_PRESENT


def _read_all_committers_from_file(path: Path) -> List[str]:
    try:
        return read_lines(path)
    except FileNotFoundError:
        return []



from typing import List

from guet.committers.committer import Committer


class CommitterPrintFormatter:

    def __str__(self):
        raise NotImplementedError


class SingleCommitterPrintFormatter(CommitterPrintFormatter):

    def __init__(self, committer: Committer):
        self._committer = committer

    def __str__(self):
        raise NotImplementedError


class MultipleCommitterPrintFormatter(CommitterPrintFormatter):
    def __init__(self, committers: List[Committer]):
        self._committers = committers

    def __str__(self):
        raise NotImplementedError



from os import mkdir
from os.path import isdir
from pathlib import Path
from typing import List

from guet import __version__, constants
from guet.config import CONFIGURATION_DIRECTORY
from guet.files import FileSystem
from guet.steps.preparation.preapration import Preparation


class InitializePreparation(Preparation):
    def __init__(self, file_system: FileSystem):
        super().__init__()
        self._file_system = file_system

    def prepare(self, args: List[str]):
        if not isdir(CONFIGURATION_DIRECTORY):
            self._create_configuration_folder()

            configuration_dir = Path(CONFIGURATION_DIRECTORY)

            self._create_empty_file(configuration_dir.joinpath(constants.COMMITTER_NAMES))
            self._create_empty_file(configuration_dir.joinpath(constants.CONFIG))
            self._create_empty_file(configuration_dir.joinpath(constants.COMMITTERS))
            self._create_empty_file(configuration_dir.joinpath(constants.COMMITTERS_SET))
            self._create_empty_file(configuration_dir.joinpath(constants.ERRORS))

            config = self._create_empty_file(Path(configuration_dir.joinpath(constants.CONFIG)))
            config.write([f'{__version__}\n', '\n'])

            self._file_system.save_all()

    def _create_empty_file(self, path: Path) -> None:
        file = self._file_system.get(path)
        file.write([])
        return file

    def _create_configuration_folder(self) -> None:
        mkdir(CONFIGURATION_DIRECTORY)



from guet.util._current_millis import current_millis
from guet.util._get_command_key import get_command_key
from guet.util._project_root import project_root
from guet.util._proxy import Proxy
from guet.util._recursive_directory_find import recursive_directory_find

from ._add_command_if_none_given import \
    add_command_help_if_invalid_command_given
from ._args import Args
from ._help_builder import FlagBuilder, FlagsBuilder, HelpMessageBuilder
